/*

Copyright 2010 Andreas Massouras

This file is part of Primer-Initiated Sequence Synthesis (PrInSeS)

For license information, please refer to the LICENSE document included in the distribution.
By downloading or using this software, you accept and are bound by the terms of the license.

*/

#include "Reads.h"
#include "Util.h"
#include <fstream>
#include <vector>
#include <iomanip>
#include <assert.h>

const char* prinsesReadFileSignature = "PrInSeS-formatted reads";
const char* prinsesReadFileVersion = "v2";

const static uint64_t lSize = sizeof(uint64_t);
const static char hexa = (char) (10 - 'a');

const static char* sequenceBinaryCodes[] = {
	"AA", "AC", "AG", "AT", "CA", "CC", "CG", "CT", 
	"GA", "GC", "GG", "GT", "TA", "TC", "TG", "TT"
};

const static int32_t mapSize = 16777216; // 24-bit index, see also index() method

static String
fromHex(const String& hex)
{
	uint64_t len = 0;
	uint64_t i;
	for (i = 0; i < lSize; i+= 2) {
		String hexlen = hex.substr(i, 2);
		unsigned int ii = 0;
		Util::value<unsigned int>(ii, hexlen, std::hex);
		len += ((uint64_t) ii) << (i*4);
	}

	const char *raw = hex.c_str();
	const char *r;
	uint64_t binlen = len / 2;
	String ret;
	for (i = 0, r = raw + lSize; i < binlen; ++i, ++r) {
		if (*r > '9')
			ret += sequenceBinaryCodes[(int) (*r + hexa)];
		else
			ret += sequenceBinaryCodes[(int) (*r - '0')];
	}
	if (len > 2 * binlen) {
		if (*r > '9')
			ret += sequenceBinaryCodes[(int) (*r + hexa)][0];
		else
			ret += sequenceBinaryCodes[(int) (*r - '0')][0];
	}
	return ret;
}


ShortSequenceSortedArray::ShortSequenceSortedArray(uint64_t number, uint16_t readLength, bool hasScores) : 
	_readLength(readLength),
	_readSize(ShortDnaBinarySequence::memorySize(readLength)),
	_scoreSize(hasScores ? (readLength+1)/2: 0),
	_recordSize(_readSize + 4 + _scoreSize), _number(0), _totalNumber(0), _map(NULL)
{
	_next = new uint8_t [_recordSize * number];
	_memory = _next;
}


void 
ShortSequenceSortedArray::add(const ShortDnaBinarySequence* s, uint32_t numCopies)
{
	((ShortDnaBinarySequence*) _next)->copy(s);
	_next += _readSize;
	* ((uint32_t*) _next) = numCopies;
	_next += 4 + _scoreSize;
	++_number;
	_totalNumber += numCopies;
}



void 
ShortSequenceSortedArray::addV1(const String& seq, uint32_t count)
{
	((ShortDnaBinarySequence*) _next)->set(seq.c_str());
	_next += _readSize;
	* ((uint32_t*) _next) = count;
	_next += 4 + _scoreSize;
	++_number;
	_totalNumber += count;
}


void 
ShortSequenceSortedArray::addV1(const String& seq, uint32_t count, const String& score)
{
	((ShortDnaBinarySequence*) _next)->set(seq.c_str());
	_next += _readSize;

	* ((uint32_t*) _next) = count;
	_next += 4;

	const char* from = score.c_str();
	uint64_t len = (score.length()+1)/2;
	uint8_t * sc = (uint8_t *) _next;
	uint8_t  c, d;
	for (uint64_t i = 0; i < len; ++i, ++sc) {
		d = from[2*i];
		c = (d > '9' ? d-'a'+10 : d-'0') * 16;
		d = from[2*i+1];
		if (d) c += d > '9' ? d-'a'+10 : d-'0';
		*sc = c;
	}
	_next += _scoreSize;
	++_number;
	_totalNumber += count;
}


ShortSequenceSortedArray*
ShortSequenceSortedArray::loadReadsV1(istream& ifile)
{
	String line, score;
	int totalNumReads = 0;
	int numReads = 0;
	uint16_t readLength = 0;
	getline(ifile, line);	// header
	vector<String> fields = Util::split(line, '\t');
	bool readsFileContainsScores;
	uint16_t readLengthDist[1024];

	if (fields.size() < 4 or !Util::value<bool>(readsFileContainsScores, fields[2])) {
		cerr << "Unrecognised reads file format"
#ifdef DEBUG
			<< " in " << __PRETTY_FUNCTION__ << " " << __FILE__ 
			<< " line " << __LINE__ 
#endif
			<< endl;
		exit(1);
	}

	for (int i = 0; i < 1024; ++i) readLengthDist[i] = 0;

	for (numReads = 0; getline(ifile, line); numReads++) {
		getline(ifile, line);
		uint16_t len = 0;
		for (uint64_t i = 0; i < lSize; i+= 2) {
			String hexlen = line.substr(i, 2);
			unsigned int ii = 0;
			Util::value<unsigned int>(ii, hexlen, std::hex);
			len += ((uint64_t) ii) << (i*4);
		}
		if (readLength < len) readLength = len;
		if (len < 1024) readLengthDist[len]++;
		if (readsFileContainsScores) getline(ifile, score);

		if (numReads % 1000000 == 0) {
			cerr << "\rLoading reads: " << setw(4) << (numReads/1000000) << " million";
		}
	}
	
	ShortSequenceSortedArray* reads = new ShortSequenceSortedArray(numReads, readLength, readsFileContainsScores);
        ifile.clear();
        ifile.seekg(0, ios::beg);       // rewind
	getline(ifile, line);	// header

	for (numReads = 0; getline(ifile, line); numReads++) {
		int num;
		if (!Util::value<int>(num, line)) {
			cerr << "Incorrect reads file format"
#ifdef DEBUG
				<< " in " << __PRETTY_FUNCTION__ 
				<< " " << __FILE__ << " line " << __LINE__
#endif
				<< endl;
			exit(1);
		}
		getline(ifile, line);
		String readSeq = fromHex(line);
		if (readsFileContainsScores) {
			getline(ifile, score);
			reads->addV1(readSeq, num, score);
		} else {
			reads->addV1(readSeq, num);
		}

		totalNumReads += num;
		if (numReads % 1000000 == 0) {
			cerr << "\rLoading reads: " << setw(4) << (numReads/1000000) << " million";
		}
	}

	cerr << "\r" << Util::getLocalTime(time(NULL)) << " found " << numReads << " unique reads, " << totalNumReads << " total, read length=" << readLength << endl << flush;
	cerr << "Read length\tNumber of reads\n";
	for (int i = 0; i < 1024; ++i) {
		if (readLengthDist[i])
			cerr << setw(4) << i << "\t" << setw(12) << readLengthDist[i] << endl;
	}
#ifdef READS_INDEXING
	reads->buildMap();
#endif
	return reads;
}


static int
compareShortDnaBinarySequences(const void* a, const void* b)
{
	return ((const ShortDnaBinarySequence*) a)->compare( (const ShortDnaBinarySequence*) b);
}


#define NEXT(bseq)  ((ShortDnaBinarySequence*) ((uint8_t*) (bseq) + recordSize))

/* static */ ShortSequenceSortedArray*
ShortSequenceSortedArray::nonUniqueShortSequences(uint8_t* memory, uint64_t n, uint16_t fragLength, ostream* logFile)
{
	if (n == 0) return NULL;
	const uint16_t recordSize = ShortDnaBinarySequence::memorySize(fragLength);
	if (logFile) *logFile << '\r' << Util::getLocalTime(time(NULL)) << " sorting genomic fragments...\n";
	qsort((void*) memory, n, recordSize, compareShortDnaBinarySequences);

	if (logFile) *logFile << Util::getLocalTime(time(NULL)) << " finding number of duplicates...\n";
	uint64_t numNonUnique = 0;
	ShortDnaBinarySequence* bseq = (ShortDnaBinarySequence*) memory;
	for (uint64_t i = 0; i < n - 1; ++i, bseq = NEXT(bseq)) {
		if (bseq->equals(NEXT(bseq))) {
			numNonUnique++;
			for (i++, bseq = NEXT(bseq);
			 i < n - 1 and bseq->equals(NEXT(bseq));
			 i++, bseq = NEXT(bseq));
		}
	}
	if (logFile) *logFile << Util::getLocalTime(time(NULL)) << " copying duplicates...\n";
	ShortSequenceSortedArray* nonUniqueShortSeq = new ShortSequenceSortedArray(numNonUnique, fragLength, false);
	bseq = (ShortDnaBinarySequence*) memory;
	for (uint64_t i = 0; i < n - 1; ++i, bseq = NEXT(bseq)) {
		if (bseq->equals(NEXT(bseq))) {
			uint32_t numCopies = 2;
			for (i++, bseq = NEXT(bseq);
			 i < n - 1 and bseq->equals(NEXT(bseq));
			 i++, bseq = NEXT(bseq))
				numCopies++;
			nonUniqueShortSeq->add(bseq, numCopies);
		}
	}

	delete [] memory;
#ifdef READS_INDEXING
	nonUniqueShortSeq->buildMap();
#endif
	return nonUniqueShortSeq;
}


#if 0
/* static */ ShortSequenceSortedArray*
ShortSequenceSortedArray::nonUniqueShortSequences(const Genome& genome, uint16_t fragLength, ostream* logFile)
{
	GenomicPosition totalNumber = 0;
	for (ChromosomeIndex i = 0; i < genome.numberChromosomes(); ++i)
		totalNumber += genome.chromoLength(i) - fragLength + 1;

	if (totalNumber == 0) {
		if (logFile) *logFile << "Empty genome!\n";
		return NULL;
	}
	const uint16_t recordSize = ShortDnaBinarySequence::memorySize(fragLength);
	uint8_t* memory = new uint8_t[(uint64_t) totalNumber * (uint64_t) recordSize * 2];
	ShortDnaBinarySequence *bseq = (ShortDnaBinarySequence*) memory;
	uint64_t allSeqsNum = 0;

	uint64_t onePercent = totalNumber / 100;
	uint64_t numDone = 1;
	if (logFile) *logFile << "Building list of fragments...\n";
	for (ChromosomeIndex i = 0; i < genome.numberChromosomes(); ++i) {
		const char* seq = genome.sequence(i, Strand::positive);
		const char* seqRC = genome.sequence(i, Strand::negative);
		uint32_t imax = genome.chromoLength(i) - fragLength;
		for (uint32_t i = 0; i <= imax; i++, numDone++) {
			if (logFile and numDone % onePercent == 0) {
				*logFile << '\r' << (numDone / onePercent) << "% done  ";
			}
			if (bseq->set(seq + i, fragLength)) {
				allSeqsNum++;
				bseq = NEXT(bseq);
			}
			if (bseq->set(seqRC + i, fragLength)) {
				allSeqsNum++;
				bseq = NEXT(bseq);
			}
		}
	}

	return nonUniqueShortSequences(memory, allSeqsNum, fragLength, logFile);
}
#endif


/* static */
ShortSequenceSortedArray*
ShortSequenceSortedArray::nonUniqueShortSequences(const DnaSequence& sequence, uint16_t fragLength, ostream* logFile)
{
	const GenomicPosition len = sequence.length();
	const uint16_t recordSize = ShortDnaBinarySequence::memorySize(fragLength);
	uint8_t* memory = new uint8_t[(uint64_t) (len - fragLength + 1) * (uint64_t) recordSize];
	ShortDnaBinarySequence *bseq = (ShortDnaBinarySequence*) memory;
	uint64_t allSeqsNum = 0;

//	uint64_t onePercent = totalNumber / 100;
//	uint64_t numDone = 1;
	if (logFile) *logFile << "Building list of fragments...\n";
	const char* seq = sequence.c_str();
//	const char* seqRC = genome.sequence(i, Strand::negative);
	uint32_t imax = len - fragLength;
	for (uint32_t i = 0; i <= imax; i++) { // , numDone++
//		if (logFile and numDone % onePercent == 0) {
//			*logFile << '\r' << (numDone / onePercent) << "% done  ";
//		}
		if (bseq->set(seq + i, fragLength)) {
			allSeqsNum++;
			bseq = NEXT(bseq);
		}
/*
		if (bseq->set(seqRC + i, length)) {
			allSeqsNum++;
			bseq = NEXT(bseq);
		}
*/
	}

	return nonUniqueShortSequences(memory, allSeqsNum, fragLength, logFile);
}


void
ShortSequenceSortedArray::buildMap()
{
	// _map[int32_t index] = uint64_t value
	if (_map) return;	// assumes it's already done

	_map = new uint64_t [mapSize];
	int32_t lastIndex = 0;
	for (uint64_t i = 0; i < _number; ++i) {
		int32_t thisIndex = index(sequence(i));
		for (; lastIndex < thisIndex; lastIndex++)
			_map[lastIndex] = i;
	}
	for (; lastIndex < mapSize; lastIndex++)
		_map[lastIndex] = _number;
}


ShortSequenceSortedArray*
ShortSequenceSortedArray::loadReads(istream& ifile)
{
	String line, score;
	uint16_t readLength = 0;
	getline(ifile, line);	// header
	vector<String> fields = Util::split(line, '\t');
	bool readsFileContainsScores;

	if (fields.size() < 4 or fields[0] != prinsesReadFileSignature
	or !Util::value<bool>(readsFileContainsScores, fields[2])) {
		cerr << "Unrecognised reads file format"
#ifdef DEBUG
			<< " in " << __PRETTY_FUNCTION__ << " " << __FILE__ 
			<< " line " << __LINE__ 
#endif
			<< endl;
		exit(1);
	}
	if (fields[1] == "v1") {
		ifile.clear(); ifile.seekg(0, ios::beg);       // rewind
		return loadReadsV1(ifile);
	}

	uint64_t numUniqueReads;
	uint64_t numReads;
	uint64_t memoryUsed;
	// header line: signature version hasScores numUniqueReads totalNumReads readLength memoryUsed localTime
	if (fields.size() < 8 or fields[1] != prinsesReadFileVersion
	or !Util::value<uint64_t>(numUniqueReads, fields[3])
	or !Util::value<uint64_t>(numReads, fields[4])
	or !Util::value<uint16_t>(readLength, fields[5])
	or !Util::value<uint64_t>(memoryUsed, fields[6])
	) {
		cerr << "Unrecognised reads file format"
#ifdef DEBUG
			<< " in " << __PRETTY_FUNCTION__ << " " << __FILE__ 
			<< " line " << __LINE__ 
#endif
			<< endl;
		exit(1);
	}

	cerr << "Loading " << numUniqueReads << " unique reads, " << numReads << " total reads, read length " << readLength << "...\n";
	ShortSequenceSortedArray* reads = new ShortSequenceSortedArray(numUniqueReads, readLength, readsFileContainsScores);

	ifile.read((char*) reads->_next, memoryUsed);
	reads->_next += memoryUsed;
	reads->_number = numUniqueReads;
	reads->_totalNumber = numReads;
#ifdef READS_INDEXING
	reads->buildMap();
#endif
	return reads;
}


/*
void
sigsegHandler(int sig)
{
	cerr << "\n" << Util::getLocalTime(time(NULL)) << " - caught signal " << sig << ", cr=" << current_read << ", hr=" << high_read << ", num reads=" << num_reads << ", maxlen=" << maxLen << endl;
	exit(1);
}
*/


// finds all reads starting with the same sequence as "frag" and are longer than frag
ShortSequenceSortedArrayIterator
ShortSequenceSortedArray::findFragment(const ShortDnaBinarySequence* frag) const
{
	uint64_t low, high;
	if (_map) {
		uint64_t ind = index(frag);
		low = ind == 0 ? 0 : _map[ind-1];
		high = _map[ind];
	} else {
		low = 0;
		high = _number;
	}
	int cmp = -1;
	uint16_t len = frag->length();

	const ShortDnaBinarySequence* readSequence = NULL;
	uint64_t current;
	while (low < high) {
		current = (low + high) / 2;
		readSequence = sequence(current);
		cmp = frag->compare(readSequence, len);
		if (cmp < 0) {
			high = current;
		} else if (cmp > 0) {
			low = current + 1;
		} else {	// cmp == 0
			break;
		}
	}

	if (cmp) return ShortSequenceSortedArrayIterator(*this);	// no reads found

	if (readSequence->length() == len) {
		low = current + 1;
		if (low >= _number or !frag->equals(sequence(low), len))
			return ShortSequenceSortedArrayIterator(*this);
		
		for (high = low + 1; high < _number and frag->equals(sequence(high), len); ++high) {}
		high--;
	} else {
		for (high = current + 1; high < _number and frag->equals(sequence(high), len); ++high) {}
		high--;
		for (low = current; low > 0; low--) {
			const ShortDnaBinarySequence* readSequence = sequence(low-1);
			if (readSequence->length() <= len or !frag->equals(readSequence, len)) break;
		}
	}

	return ShortSequenceSortedArrayIterator(*this, low, high);
}


// finds all reads whose first n bases are the same as those of "frag"
ShortSequenceSortedArrayIterator
ShortSequenceSortedArray::findFragment(const ShortDnaBinarySequence* frag, uint16_t n) const
{
	uint64_t low, high;
	if (_map) {
		uint64_t ind = index(frag);
		low = ind == 0 ? 0 : _map[ind-1];
		high = _map[ind];
	} else {
		low = 0;
		high = _number;
	}

	int cmp = -1;
	assert (n <= frag->length());

	const ShortDnaBinarySequence* readSequence = NULL;
	uint64_t current;
	while (low < high) {
		current = (low + high) / 2;
		readSequence = sequence(current);
		cmp = frag->compare(readSequence, n);
		if (cmp < 0) {
			high = current;
		} else if (cmp > 0) {
			low = current + 1;
		} else {	// cmp == 0
			break;
		}
	}

	if (cmp) return ShortSequenceSortedArrayIterator(*this);	// no reads found

	for (high = current; high + 1 < _number and frag->equals(sequence(high + 1), n); ++high) {}
	for (low = current; low > 0 and frag->equals(sequence(low-1), n); low--);

	return ShortSequenceSortedArrayIterator(*this, low, high);
}


bool
ShortSequenceSortedArray::contains(const ShortDnaBinarySequence* frag) const
{
	uint16_t len = frag->length();
	if (len > _readLength) return false;

	uint64_t low, high;
	if (_map) {
		uint64_t ind = index(frag);
		low = ind == 0 ? 0 : _map[ind-1];
		high = _map[ind];
	} else {
		low = 0;
		high = _number;
	}

	int cmp = -1;


	while (low < high) {
		uint64_t current = (low + high) / 2;
		const ShortDnaBinarySequence* readSequence = sequence(current);
		cmp = frag->compare(readSequence, len);
		if (cmp < 0) {
			high = current;
		} else if (cmp > 0) {
			low = current + 1;
		} else {	// cmp == 0
			return true;
		}
	}

	return false;
}
