/*

Copyright 2010 Andreas Massouras

This file is part of Primer-Initiated Sequence Synthesis (PrInSeS)

For license information, please refer to the LICENSE document included in the distribution.
By downloading or using this software, you accept and are bound by the terms of the license.

*/


#include "Threads.h"

using namespace std;

void *
ThreadScheduler::runOneJob(void* j)
{
	ThreadedJob* job = (ThreadedJob*) j;
	job->run();
	job->_scheduler->signal(job);
	return NULL;
}


ThreadScheduler::ThreadScheduler(uint16_t maxThreads)
	: _numThreadsRunning(0), _runningJobs(maxThreads, NULL)
{
	pthread_mutex_init(&_mutex, NULL);
	pthread_cond_init(&_cond, NULL);
}


ThreadScheduler::~ThreadScheduler()
{
	assert(_jobQueue.size() == 0);
	assert(_numThreadsRunning == 0);
	pthread_mutex_destroy(&_mutex);
	pthread_cond_destroy(&_cond);
}


void
ThreadScheduler::addJob(ThreadedJob* job)
{
//	cdbg << "addJob in\n";
	job->_scheduler = this;
	_jobQueue.push_back(job);
	runNextJob();
//	cdbg << "addJob out\n";
}


ThreadedJob*
ThreadScheduler::waitForNext()
{
//	cdbg << "waiting for next...\n";
	pthread_mutex_lock(&_mutex);
//	cdbg << "num running " << _numThreadsRunning << " finished " << _jobsFinished.size() << " queued " << _jobQueue.size() << endl;
	if (_jobsFinished.size()) {
		ThreadedJob* j = _jobsFinished.front();
		_jobsFinished.pop_front();
		pthread_mutex_unlock(&_mutex);
		return j;
	}
	if (_numThreadsRunning == 0) {
		pthread_mutex_unlock(&_mutex);
		if (_jobQueue.empty()) return NULL;
		runNextJob();
		pthread_mutex_lock(&_mutex);
	}

	pthread_cond_wait(&_cond, &_mutex);
//	cdbg << "got cond wait returned\n";

	ThreadedJob* j = _jobsFinished.front();
	_jobsFinished.pop_front();

	// this should return immediately:
	while (_threadsFinished.size()) {
//		cdbg << "joining thread\n";
		pthread_join(_threadsFinished.front(), NULL);
		_threadsFinished.pop_front();
	}
//	cdbg << "all threads joined\n";
//	_numThreadsRunning--;
	pthread_mutex_unlock(&_mutex);

	runNextJob();
	return j;
}


// runs in job's thread:
void
ThreadScheduler::signal(ThreadedJob* job)
{
//	cdbg << "running signal in\n";
	pthread_mutex_lock(&_mutex);
	_jobsFinished.push_back(job);
	uint16_t i;
	for (i = 0; i < _runningJobs.dimension() and _runningJobs[i] != job; ++i);
	assert (i < _runningJobs.dimension());
	_threadsFinished.push_back(job->_thread);
	_runningJobs[i] = NULL;
	assert(_numThreadsRunning);
	_numThreadsRunning--;
	pthread_cond_signal(&_cond);
	pthread_mutex_unlock(&_mutex);
}


void
ThreadScheduler::runNextJob()
{
	pthread_mutex_lock(&_mutex);

	while (!_jobQueue.empty() and _numThreadsRunning < _runningJobs.dimension()) {
		// find empty slot:
		uint16_t i;
		for (i = 0; i < _runningJobs.dimension() and _runningJobs[i]; ++i);

		assert (i < _runningJobs.dimension());

		_runningJobs[i] = _jobQueue.front();
		_jobQueue.pop_front();
	//	cdbg << "starting thread " << i << "\n";
		pthread_create(&_runningJobs[i]->_thread, NULL, runOneJob, (void*) _runningJobs[i]);
		_numThreadsRunning++;
	}

	pthread_mutex_unlock(&_mutex);
}
