/*

Copyright 2010 Andreas Massouras

This file is part of Primer-Initiated Sequence Synthesis (PrInSeS)

For license information, please refer to the LICENSE document included in the distribution.
By downloading or using this software, you accept and are bound by the terms of the license.

*/

#ifndef THREADS_H
#define THREADS_H

#include <pthread.h>
#include <list>
#include "Vector.h"


class Mutex
{
public:
	Mutex() { pthread_mutex_init(&_mutex, NULL); }
	~Mutex() { pthread_mutex_destroy(&_mutex); }
	void lock() { pthread_mutex_lock(&_mutex); }
	void unlock() { pthread_mutex_unlock(&_mutex); }

private:
	pthread_mutex_t _mutex;
};


class MutexLock
{
public:
	MutexLock(Mutex& m) : _mutex(m) { m.lock(); }
	~MutexLock() { _mutex.unlock(); }

private:
	Mutex& _mutex;
};


class ThreadScheduler;

class ThreadedJob
{
public:

	virtual void run() = 0;
	ThreadedJob() : _scheduler(NULL) {}
	virtual ~ThreadedJob() {}

protected:
	ThreadScheduler* _scheduler;

private:
	pthread_t _thread;

	friend class ThreadScheduler;
};

// Not sure if it's a good idea to instantiate more than one of these:
class ThreadScheduler
{
public:
	friend class ThreadedJob;

	// the following all run in main thread:
	ThreadScheduler(uint16_t maxThreads);
	~ThreadScheduler();
	void addJob(ThreadedJob* job);
	ThreadedJob* waitForNext();
	void waitForAll() { while(waitForNext()); }
	uint16_t numThreadsRunning() const { return _numThreadsRunning; }
	uint32_t numJobsQueued() const { return _jobQueue.size(); }
	uint32_t numJobsFinished() const { return _jobsFinished.size(); }

private:
	pthread_mutex_t _mutex;
	pthread_cond_t _cond;

	uint16_t _numThreadsRunning;
	list<ThreadedJob*> _jobQueue;
	list<ThreadedJob*> _jobsFinished;
	list<pthread_t> _threadsFinished;
	Vector<ThreadedJob*> _runningJobs;

	void signal(ThreadedJob*);	// signal job's done, runs in job's thread
	static void* runOneJob(void*);	// runs in job's thread
	void runNextJob();		// runs in main thread
};

#endif // #ifndef THREADS_H
