/*

Copyright 2010 Andreas Massouras

This file is part of Primer-Initiated Sequence Synthesis (PrInSeS)

For license information, please refer to the LICENSE document included in the distribution.
By downloading or using this software, you accept and are bound by the terms of the license.

*/

#include <iomanip>
#include <fstream>
#include <stdio.h>
#include <sys/stat.h>
#include <assert.h>
#include "Util.h"
#include "Vector.h"
#include <unistd.h>
#include <limits>

// #define DEBUG_UTIL

ostream& cdbg = cerr;

const GenomicPosition invalidGenomicPosition = numeric_limits<GenomicPosition>::max();

namespace Util {

static ostream& cdbg = cerr;

vector<String>
getCommandLineOptions(const map<String, bool*>& flags, const map<String, int*> intOptions, const map<String, String*>& stringOptions, int argc, char* argv[], void (*usage)())
{
	vector<String> ret;

	for (int i = 1; i < argc; ++i) {
		int arglen = strlen(argv[i]);
		if (arglen <= 1 or argv[i][0] != '-') {
			// regular argument, put it in the returned vector
			ret.push_back(String(argv[i]));
#ifdef DEBUG_UTIL
			cerr << argv[i] << " regular argument\n";
#endif
			continue;
		}

		const char *thisArg = argv[i]+1;
		arglen--;
		if (arglen > 1 and thisArg[0] == '-') {
			thisArg++;
			arglen--;
		}
		bool found = false;
		map<String, bool*>::const_iterator bi;
		for (bi = flags.begin(); bi != flags.end(); ++bi) {
			if ((*bi).first.compare(thisArg) == 0) {
				*((*bi).second) = true;
				found = true;
#ifdef DEBUG_UTIL
				cerr << argv[i] << " bool (true)\n";
#endif
				break;
			}
			if (arglen > 2 and thisArg[0] == 'n' and thisArg[1] == 'o'
			and (*bi).first.compare(thisArg+2) == 0) {
				*((*bi).second) = false;
				found = true;
#ifdef DEBUG_UTIL
				cerr << argv[i] << " bool (false)\n";
#endif
				break;
			}
		}
		if (found) continue;
		map<String, int*>::const_iterator ii;
		for (ii = intOptions.begin(); ii != intOptions.end(); ++ii) {
			if ((*ii).first.compare(thisArg) == 0) {
				if (i == argc-1
				or sscanf(argv[i+1], "%d", (*ii).second) != 1) {
					cerr << "Integer value expected for option " << argv[i] << endl;
					usage();
					exit(1);
				}

#ifdef DEBUG_UTIL
				cerr << argv[i] << " int (" << *((*ii).second) << ")\n";
#endif
				found = true;
				i++;
				break;
			}
		}
		if (found) continue;

		map<String, String*>::const_iterator oi;
		for (oi = stringOptions.begin(); oi != stringOptions.end(); ++oi) {
			if ((*oi).first.compare(thisArg) == 0) {
				if (i == argc-1) {
					cerr << "Value expected for option " << argv[i] << endl;
					usage();
					exit(1);
				}
				*((*oi).second) = argv[i+1];
#ifdef DEBUG_UTIL
				cerr << argv[i] << " string (" << *((*oi).second )<< ")\n";
#endif
				found = true;
				i++;
				break;
			}
		}
		if (found) continue;
		cerr << "Unknown option " << argv[i] << endl;
		usage();
		exit(1);
	}

	return ret;
}


String
hms(time_t seconds)
{
	int sec = seconds % 60;
	int minutes = (seconds - sec) / 60;
	int min = minutes % 60;
	int hours = (minutes - min) / 60;

	stringstream ss;
	ss << hours 
		<< ":" << setfill('0') << setw(2) << min 
		<< ":" << setfill('0') << setw(2) << sec;
	return ss.str();
}


int
split(const String& str, String *fields[], char separator)
{
	uint64_t pos = 0;
	uint64_t end;
	int numFields = 0;

	while (*fields) {
		end = str.find(separator, pos);
		**fields = str.substr(pos, end-pos);
		numFields++;
		fields++;
		if (end == string::npos) break;
		pos = end+1;
	}

	while (*fields) {
		(**fields).clear();
		fields++;
	}
	
	return numFields;
}


void
split(const String& str, const char separator, /* returned */ Vector<String>& fields)
{
	const char* s = str.c_str();
	if (!*s) {
		fields.resize(0);
		return;
	}

	uint32_t numFields = 1;
	for (; *s; ++s)
		if (*s == separator) numFields++;

	fields.resize(numFields);
	const char *from, *to;
	uint32_t fieldNo = 0;
	for (from = to = str.c_str(); *to; ++to) {
		if (*to == separator) {
			fields[fieldNo++].assign(from, (size_t) (to - from));
			from = to + 1;
		}
	}
	fields[fieldNo++].assign(from, (size_t) (to - from));
	assert(fieldNo == numFields);
}


String
reverseComplement(const String& seq)
{
	String ret(seq);
	uint64_t i, j;
	for (i = seq.length(), j = 0; i > 0; --i, ++j) {
		switch(seq[i-1]) {
		case 'A': ret[j] = 'T'; break;
		case 'C': ret[j] = 'G'; break;
		case 'G': ret[j] = 'C'; break;
		case 'T': ret[j] = 'A'; break;
		case 'a': ret[j] = 't'; break;
		case 'c': ret[j] = 'g'; break;
		case 'g': ret[j] = 'c'; break;
		case 't': ret[j] = 'a'; break;
		case 'R': ret[j] = 'Y'; break;
		case 'Y': ret[j] = 'R'; break;
		case 'M': ret[j] = 'K'; break;
		case 'K': ret[j] = 'M'; break;
		case 'r': ret[j] = 'y'; break;
		case 'y': ret[j] = 'r'; break;
		case 'm': ret[j] = 'k'; break;
		case 'k': ret[j] = 'm'; break;
		default: ret[j] = seq[i-1];
		}
	}
	return ret;
}


void
reverseComplement(char* to, const char* from, uint32_t n)
{
	const char* c = from + n - 1;
	for (; c >= from; c--, to++) {
		switch(*c) {
		case 'A': *to = 'T'; break;
		case 'C': *to = 'G'; break;
		case 'G': *to = 'C'; break;
		case 'T': *to = 'A'; break;
		case 'a': *to = 't'; break;
		case 'c': *to = 'g'; break;
		case 'g': *to = 'c'; break;
		case 't': *to = 'a'; break;
		case 'R': *to = 'Y'; break;
		case 'Y': *to = 'R'; break;
		case 'M': *to = 'K'; break;
		case 'K': *to = 'M'; break;
		case 'r': *to = 'y'; break;
		case 'y': *to = 'r'; break;
		case 'm': *to = 'k'; break;
		case 'k': *to = 'm'; break;
		default: *to = *c;
		}
	}
	*to = 0;
}


int
parseTabSeparatedFile(const char*& str, string *fields[])
{
	if (!*str) return 0;

	const char* end = str;
	int numFields = 0;
	while (1) {
		if (!*end or *end == '\n') {
			if (*fields) {
				**fields = string(str, end-str);
				fields++;
			}
			str = end + 1;
			numFields++;
			break;
		}
		if (*end == '\t') {
			if (*fields) {
				**fields = string(str, end-str);
				fields++;
			}
			str = end + 1;
			numFields++;
		}
		end++;
	}
	while (*fields) {
		(**fields).clear();
		fields++;
	}
	return numFields;
}


uint64_t
getMemoryUsage()
{
	std::ifstream stat1("/proc/self/stat");
	if (stat1.fail()) return 0;

	String line;
	getline(stat1, line);
	Vector<String> stuff;
	split(line, ' ', stuff);
	uint64_t ret;
	value(ret, stuff[22]);
	return ret;
}


void
checkCreateDirectory(const char* name)
{
	struct stat st;
	if (!stat(name, &st)) {
		if (!S_ISDIR(st.st_mode)) {
			cerr << "File " << name << " exists but is not a directory"
#ifdef DEBUG  
				<< " in " << __PRETTY_FUNCTION__ 
				<< " " << __FILE__ << " line " << __LINE__ 
#endif
				<< endl;
			exit(1);
		}
		return;
	}
	
	// create the directory
	if (mkdir(name, 0777) != 0) {
		cerr << "Can't create directory " << name 
#ifdef DEBUG  
			<< " in " << __PRETTY_FUNCTION__ << " " << __FILE__ 
			<< " line " << __LINE__ 
#endif
			<< endl;
		exit(1);
	}
}


String
getCurrentDirectory()
{
	char buffer[FILENAME_MAX];
	assert(getcwd(buffer, sizeof(buffer)));
	return String(buffer);
}


// This makes numbers show with thousand separator
struct LocaleWithThousandSeparator : public std::numpunct<char>
{
protected:
	string do_grouping () const {return "\003"; }
	char_type do_thousands_sep() const { return '\''; }
};

static const LocaleWithThousandSeparator lwts;


void
setThousandSeparator(ostream& o)
{
	o.imbue( locale( o.getloc(), &lwts));
}


bool
valueI(int32_t& i, const String& s)
{
	const char* c = s.c_str();
	if (!*c) return false;
	i = 0;
	bool negative = false;
	while (*c == ' ') c++;
	if (*c == '-') {
		negative = true;
		c++;
	}
	while (*c == ' ') c++;
	for (; *c >= '0' and *c <= '9'; c++) {
		i *= 10;
		i += (*c - '0');
	}
	while (*c == ' ') c++;
	if (*c) return false;
	if (negative) i = -i;
	return true;
}


bool
valueU(uint32_t& i, const String& s)
{
	const char* c = s.c_str();
	if (!*c) return false;
	i = 0;
	while (*c == ' ') c++;
	for (; *c >= '0' and *c <= '9'; c++) {
		i *= 10;
		i += (*c - '0');
	}
	while (*c == ' ') c++;
	if (*c) return false;
	return true;
}


} // namespace Util


uint8_t** SimpleAllocator::_blocks;
uint32_t SimpleAllocator::_numberBlocks;
uint32_t SimpleAllocator::_currentBlock;
uint8_t* SimpleAllocator::_next;


uint8_t*
SimpleAllocator::allocate(uint32_t size)
{
#ifdef DEBUG
	return new uint8_t [size];
#else
	if (size > 65536) {
		cerr << "Large size (" << size << ") for SimpleAllocator\n";
		throw("Large size for SimpleAllocator");
	}
	if (_next + size <= _blocks[_currentBlock] + _arraySize) {
		uint8_t* ret = _next;
		_next += size;
		return ret;
	}
	_currentBlock++;
	if (_currentBlock == _numberBlocks) {
		_numberBlocks *= 2;
		uint8_t** newBlocks = new uint8_t* [_numberBlocks];
		for (uint16_t i = 0; i < _currentBlock; ++i)
			newBlocks[i] = _blocks[i];
		delete [] _blocks;
		_blocks = newBlocks;
	}
	uint8_t* ret = _blocks[_currentBlock] = new uint8_t[_arraySize];
	_next = ret + size;
	return ret;
#endif
}
