/*

Copyright 2010 Andreas Massouras

This file is part of Primer-Initiated Sequence Synthesis (PrInSeS)

For license information, please refer to the LICENSE document included in the distribution.
By downloading or using this software, you accept and are bound by the terms of the license.

*/

#include "Vector.h"

static const uint32_t bubbleSortThreshold = 4;

namespace VectorLocal {	// to avoid name clashes
class LowHigh 
{
public:
	uint32_t _low;
	uint32_t _high;
	LowHigh(uint32_t l, uint32_t h) : _low(l), _high(h) {}
	LowHigh() {}
	uint32_t middle() const { return (_low + _high) / 2; }
};
};


// sorts and returns a vector r where this->[r[i]] is sorted for 0 <= i < dimension
// uses quicksort
VectorI
VectorD::orderQSort() const
{
	const uint32_t n = _dimension;
	if (n <= 1) return VectorI(n, 0);

	VectorI r(n);
	for (uint32_t i = 0; i < n; ++i) r[i] = i;

	Stack<VectorLocal::LowHigh> stack;
	stack.push(VectorLocal::LowHigh(0, n-1));

#ifdef DEBUG_SORT
	uint64_t numComparisons = 0;
	uint64_t nc1 = 0, nc2 = 0, nc3 = 0, nc4 = 0;
#endif
	while (!stack.isEmpty()) {
		VectorLocal::LowHigh lh = stack.pop();
//		cdbg << "QQLH=" << lh._low << ", " << lh._high << endl;

		// trivial if only a few elements:
		if (lh._low + bubbleSortThreshold >= lh._high) {
			while (lh._low < lh._high) {
				uint32_t min = lh._low;
				for (uint32_t i = lh._low + 1; i <= lh._high; ++i) {
#ifdef DEBUG_SORT
					numComparisons++; nc1++;
#endif
					if (_array[r[i]] < _array[r[min]]) min = i;
				}
				if (min != lh._low) r.swap(min, lh._low);
				lh._low++;
			}
			continue;
		}

		// partition:
		// first put the median of (low, mid, high) at the middle as a pivot:
		const uint32_t mid = lh.middle();
#ifdef DEBUG_SORT
		numComparisons++; nc2++;
#endif
		if (_array[r[mid]] < _array[r[lh._low]]) {
			r.swap(mid, lh._low);
		}
#ifdef DEBUG_SORT
		numComparisons++; nc2++;
#endif
		if (_array[r[lh._high]] < _array[r[mid]]) {
			r.swap(mid, lh._high);
#ifdef DEBUG_SORT
			numComparisons++; nc2++;
#endif
			if (_array[r[mid]] < _array[r[lh._low]]) {
				r.swap(mid, lh._low);
			}
		}

		const uint32_t pivotIndex = r[mid];
		const double pivotValue = _array[pivotIndex];
		// at this stage _array[r[lh._low]] < _array[pivotIndex] and _array[r[lh._high]] > _array[pivotIndex]
		r[mid] = r[lh._high-1];
#ifdef SORT_METHOD_1		// "collapsing walls" method:
		uint32_t low = lh._low + 1;
		uint32_t high = lh._high - 2;
		while (low < high) {
			while (low < high and (
#ifdef DEBUG_SORT
			++numComparisons, ++nc3,
#endif
			_array[r[low]] < pivotValue))
				low++;
			while (low < high and (
#ifdef DEBUG_SORT
			++numComparisons, ++nc3,
#endif
			pivotValue < _array[r[high]]))
				high--;

			if (low < high) {
				r.swap(low, high);
				low++;
				high--;
			}
		}
#ifdef DEBUG_SORT
		numComparisons++; nc4++;
#endif
		const uint32_t pivot = _array[r[low]] < pivotValue ? low + 1 : low;

#else	// method 2 - this can be disastrously slow in some cases
		uint32_t pivot = lh._low + 1;
		for (uint32_t i = lh._low + 1; i < lh._high - 1; ++i) {
#ifdef DEBUG_SORT
			numComparisons++; nc3++;
#endif
			if (_array[r[i]] < pivotValue) {
				if (i != pivot) r.swap(i, pivot);
				pivot++;
			}
		}
#endif	 // SORT_METHOD_1

		if (pivot > lh._low + 1)
			stack.push(VectorLocal::LowHigh(lh._low, pivot-1));

		if (pivot + 1 < lh._high) {
			r[lh._high-1] = r[pivot];
			stack.push(VectorLocal::LowHigh(pivot+1, lh._high));
		}
		r[pivot] = pivotIndex;
	}

#ifdef DEBUG_SORT
	cdbg << "Number of comparisons " << numComparisons
		<< " nc1=" << nc1 << " nc2=" << nc2 << " nc3=" << nc3 << " nc4=" << nc4
		<< " number of elements " << n
		<< " N log(N) " << n * (uint64_t) log2(n)
		<< endl;
#endif

	return r;
}


VectorI
VectorD::orderMSort() const
{
	const uint32_t n = _dimension;
	if (n <= 1) return VectorI(n, 0);

	int32_t * tmp = new int32_t [n];
	VectorI ret(n);
	for (uint32_t i = 0; i < n; ++i) ret[i] = i;

#ifdef DEBUG_SORT
	uint32_t numComp = 0;
	msort(ret._array, tmp, n, numComp);
	cdbg << "Number of comparisons " << numComp
		<< " number of elements " << n
		<< " N log(N) " << n * (uint64_t) log2(n)
		<< endl;

#else
	msort(ret._array, tmp, n);
#endif

	delete [] tmp;
	return ret;
}


static int
compareI(const void* a, const void* b)
{
	return *((const int32_t*) a) - *((const int32_t*) b);
}


void
VectorI::sort()
{
	if (_dimension > 1)
		qsort(_array, _dimension, sizeof(int32_t), compareI);
}


static int
compareD(const void* a, const void* b)
{
	if (*((const double*) a) < *((const double*) b)) return -1;
	if (*((const double*) a) > *((const double*) b)) return 1;
	return 0;
}


void
VectorD::sort()
{
	if (_dimension > 1)
		qsort(_array, _dimension, sizeof(double), compareD);
}


void
VectorD::msort(int32_t* const ret, int32_t* const tmp, const uint32_t n
#ifdef DEBUG_SORT
	, uint32_t& numComp
#endif
) const
{
	if (n <= 1) return;
	uint32_t n1 = n / 2;
	uint32_t n2 = n - n1;
	int32_t* b1 = ret;
	int32_t* b2 = ret + n1;
#ifdef DEBUG_SORT
	msort(b1, tmp, n1, numComp);
	msort(b2, tmp, n2, numComp);
#else
	msort(b1, tmp, n1);
	msort(b2, tmp, n2);
#endif

	// merge results
	int32_t* t;
	for (t = tmp; n1 and n2; t++) {
#ifdef DEBUG_SORT
		numComp++;
#endif
		if (_array[*b1] <= _array[*b2]) {
			*t = *b1;
			b1++;
			n1--;
		} else {
			*t = *b2;
			b2++;
			n2--;
		}
	}
	if (n1) memcpy(t, b1, n1 * sizeof(int32_t));
	memcpy(ret, tmp, (n - n2) * sizeof(int32_t));
}


// returns a vector r where each element is the rank of *this.
// for each i, j (0 <= i, j < n) if r[i] < r[j] then this->[i] < this->[j]
// rank starts with 1
VectorI
VectorD::rank() const
{
	VectorI o = order();
	const uint32_t n = _dimension;
	VectorI r(n);
	for (uint32_t i = 0; i < n; ++i)
		r[o[i]] = i + 1;
	return r;
}

void
VectorI::meanAndVariace(/* returned */ double& mean_, double &variance_) const
{
	const uint32_t n = _dimension;
	assert(n > 1);
	double sumsq = 0.0;
	mean_ = mean();
	for (uint32_t i = 0; i < n; ++i) {
		const double v = (double) _array[i] - mean_;
		sumsq += v * v;
	}
	variance_ = sumsq / (double) (n - 1);
}


void
VectorD::meanAndVariace(/* returned */ double& mean_, double &variance_) const
{
	const uint32_t n = _dimension;
	assert(n > 1);
	double sumsq = 0.0;
	mean_ = mean();
	for (uint32_t i = 0; i < n; ++i) {
		const double v = element(i) - mean_;
		sumsq += v * v;
	}
	variance_ = sumsq / (double) (n - 1);
}


VectorD
VectorD::averageRank() const
{
	VectorI o = order();
	const uint32_t n = _dimension;
	VectorD ar(n);
	for (uint32_t i = 0; i < n; i++) {
		uint32_t ord = o[i];
		uint32_t sumRank = i + 1;
		uint32_t num = 1;
		uint32_t j = i;
		while (j) {
			j--;
			if (element(o[j]) == element(ord)) {
				sumRank += j + 1;
				num++;
			}
		}
		const double averageRank = ((double) sumRank) / num;
		for (j = i; num; j--, num--) {
			ar[o[j]] = averageRank;
		}
	}
	return ar;
}


// sorts and returns a vector r where this->[r[i]] is sorted for 0 <= i < dimension
// uses quicksort
VectorI
VectorI::order() const
{
	if (_dimension <= 1) return VectorI(_dimension, 0);

	VectorI r(_dimension);
	for (uint32_t i = 0; i < _dimension; ++i) r[i] = i;

	Stack<VectorLocal::LowHigh> stack;
	stack.push(VectorLocal::LowHigh(0, _dimension-1));

	while (!stack.isEmpty()) {
		VectorLocal::LowHigh lh = stack.pop();

		// trivial if only a few elements:
		if (lh._low + bubbleSortThreshold >= lh._high) {
			while (lh._low < lh._high) {
				uint32_t min = lh._low;
				for (uint32_t i = lh._low + 1; i <= lh._high; ++i) {
					if (_array[r[i]] < _array[r[min]]) min = i;
				}
				if (min != lh._low) r.swap(min, lh._low);
				lh._low++;
			}
			continue;
		}

		// partition:
		// first put the median of (low, mid, high) at the middle as a pivot:
		const uint32_t mid = lh.middle();
		if (_array[r[mid]] < _array[r[lh._low]]) {
			r.swap(mid, lh._low);
		}
		if (_array[r[lh._high]] < _array[r[mid]]) {
			r.swap(mid, lh._high);
			if (_array[r[mid]] < _array[r[lh._low]]) {
				r.swap(mid, lh._low);
			}
		}

		const uint32_t pivotIndex = r[mid];
		const int32_t pivotValue = _array[pivotIndex];
		// at this stage _array[r[lh._low]] < _array[pivotIndex] and _array[r[lh._high]] > _array[pivotIndex]
		r[mid] = r[lh._high-1];
#ifdef SORT_METHOD_1		// "collapsing walls" method:
		uint32_t low = lh._low + 1;
		uint32_t high = lh._high - 2;
		while (low < high) {
			while (low < high and _array[r[low]] < pivotValue)
				low++;
			while (low < high and pivotValue < _array[r[high]])
				high--;
			if (low < high) {
				r.swap(low, high);
				low++;
				high--;
			}
		}
		const uint32_t pivot = _array[r[low]] < pivotValue ? low + 1 : low;

#else	// method 2

		uint32_t pivot = lh._low + 1;
		for (uint32_t i = lh._low + 1; i < lh._high - 1; ++i) {
			if (_array[r[i]] < pivotValue) {
				if (i != pivot) r.swap(i, pivot);
				pivot++;
			}
		}
#endif

		if (pivot > lh._low + 1)
			stack.push(VectorLocal::LowHigh(lh._low, pivot-1));

		if (pivot + 1 < lh._high) {
			r[lh._high-1] = r[pivot];
			stack.push(VectorLocal::LowHigh(pivot+1, lh._high));
		}
		r[pivot] = pivotIndex;
	}
	return r;
}


// same as rank() but averages for equal values
VectorD
VectorI::averageRank() const
{
	VectorI o = order();
	const uint32_t n = _dimension;
	VectorD ar(n);
	for (uint32_t i = 0; i < n; i++) {
		uint32_t ord = o[i];
		uint32_t sumRank = i + 1;
		uint32_t num = 1;
		uint32_t j = i;
		while (j) {
			j--;
			if (_array[o[j]] == _array[ord]) {
				sumRank += j + 1;
				num++;
			}
		}
		const double averageRank = ((double) sumRank) / num;
		for (j = i; num; j--, num--) {
			ar[o[j]] = averageRank;
		}
	}
	return ar;
}


uint32_t
VectorD::findSorted(double d) const
{
	if (_dimension == 0) return 0;
	uint32_t low = 0, high = _dimension;
	while (low < high) {
		const uint32_t middle = (low + high) / 2;
		if (d < _array[middle]) {
			high = middle;
		} else {
			low = middle + 1;
		}
	}
	if (low == _dimension or d < _array[low]) return low;
	return low + 1;
}
