/* 

Copyright 2010 Andreas Massouras

This file is part of Primer-Initiated Sequence Synthesis (PrInSeS)

For license information, please refer to the LICENSE document included in the distribution.
By downloading or using this software, you accept and are bound by the terms of the license.

*/


#include <iomanip>
#include <fstream>
#include <signal.h>
#include <limits>
#include "Prinses.h"
#include "Variant.h"
#include "Threads.h"
#include "SamPileup.h"

using namespace std;

namespace PrinsesG
{

static String programName;

// Command line arguments and their default values:
static bool printVersion = false;
static int minMatchLength = 20;
static int minDepth = 3;
static int minMappingQuality = 20;
static int maxDepth = 10;
static int rollingDepthLow = 0;
static int readDepthLow = 0;
static int readDepthHigh = 0;
static int maxTerminatorDistance = 10000;
static int maxInsertionLength = 10000;
static ostream* viewFile = NULL;
static int terminatorLength = 0;
static int primerLength = 0;
static bool verbose = false;
static bool useScores = true;
static int maxThreads = sysconf(_SC_NPROCESSORS_ONLN);
static int numTerminators = 5;
static const int maxBranches = 0;
static bool allowRepeatsInReference = false;
static bool showProgress = false;

// Global variables:
Mutex variantsRawFileMutex;
ofstream* variantsRawFile = NULL;
Mutex cerrMutex;

static void
usage()
{
	cerr << "Usage: " << programName << " [options] genome.fa alignments.bam reads.prinses prinsesDirectory" << endl;
	cerr << "\talignments.bam: sorted, indexed bam file containing all read alignments" << endl;
	cerr << "Options:" << endl;
	cerr << "  -version                 Prints version and exits" << endl;
	cerr << "  -minMatchLength n        The minimum length for matching (default " << minMatchLength << ")" << endl;
	cerr << "  -minDepth n              Minimum read depth for assembly to continue (default " << minDepth << ")" << endl;
	cerr << "  -maxDepth n              Read depth where assembly stops looking for more reads (default " << maxDepth << ")" << endl;
	cerr << "  -readDepthLow n          Read depth reported by alignment program below which assembly is always attempted (default " << readDepthLow << ")" << endl;
	cerr << "  -readDepthHigh n         Read depth reported by alignment program above which assembly is never attempted (default 1.5 times average read depth)" << endl;
	cerr << "  -maxInsertionLength n    Maximum size for insertions or deletions (default " << maxInsertionLength << ")" << endl;
	cerr << "  -maxTerminatorDistance n Maximum distance of terminator from primer (default " << maxTerminatorDistance << ")" << endl;
	cerr << "  -viewFile filename       Generate file containing one line per position" << endl;
	cerr << "  -verbose                 More information in the viewFle" << endl;
	cerr << "  -terminatorLength        Length of assembly terminator (default read length - 1)" << endl;
	cerr << "  -noscores                Ignore quality scores" << endl;
	cerr << "  -maxThreads              Maximum number of threads allowed at a time (default: number of CPUs)" << endl;
//	cerr << "  -uniquePrimers           Require primers and terminators to be unique in the genome" << endl;
	cerr << "  -numTerminators          Number of consecutive terminators at each assembly point (default " << numTerminators << ")" << endl;
	cerr << "  -allowRepeatsInReference Accept contigs where the corresponding fragment of the reference sequence constains repeats" << endl;
	cerr << "  -minMappingQuality n     When calling SNPs, ignore alignments with mapping quality less than n (default " << minMappingQuality << ")" << endl;
//	cerr << "  -maxBranches             Maximum number of branches allowed (use for heterozygous genomes)" << endl;
	cerr << "  -progress                Display progress" << endl;

	exit(1);
}


static void
signalHandler(int sig)
{
	cerr << endl << Util::getLocalTime(time(NULL)) << " " << programName << " - caught signal " << sig << ", exiting" << endl;
	exit(1);
}

#if 0
static void
breakDownVariant(bool reverse, DnaSequence oldseq, const PrinsesVariant& thisVariant, /* appends to: */ list<PrinsesVariant>& variants)
{
	DnaSequence newseq(thisVariant.newSeq());
	int pos = thisVariant.position();
	int oldseqLength, newseqLength;

	while ((oldseqLength = oldseq.length()) and (newseqLength = newseq.length())) {
		int length = oldseqLength < newseqLength ? oldseqLength : newseqLength;
		int i;
		if (reverse) {
			for (i = 0; i + 1 < length and !oldseq.contains(newseq, oldseqLength - i - 2, 2, newseqLength - i - 2, 2); ++i);
		} else {
			for (i = 0; i + 1 < length and !oldseq.contains(newseq, i, 2, i, 2); ++i);
		}

		if (i + 1 == length) {
			variants.push_back(PrinsesVariant(pos, oldseqLength, newseq, thisVariant));
			return;
		}
		
		if (reverse) {
			if (i > 0) {
				variants.push_back(PrinsesVariant(pos + oldseqLength - i, i, newseq.substr(newseqLength - i, i), thisVariant));
			}
			for (i += 2; i < length and oldseq[oldseqLength-i-1] == newseq[newseqLength-i-1]; ++i);
			oldseq = oldseq.substr(0, oldseqLength - i);
			newseq = newseq.substr(0, newseqLength - i);
		} else {
			if (i > 0) {
				variants.push_back(PrinsesVariant(pos, i, newseq.substr(0, i), thisVariant));
			}
			for (i += 2; i < length and oldseq[i] == newseq[i]; ++i);
			oldseq = oldseq.substr(i);
			newseq = newseq.substr(i);
			pos += i;
		}
	}

	if (oldseq.length()) {
		variants.push_back(PrinsesVariant(pos, oldseq.length(), DnaSequence(), thisVariant));
	} else if (newseq.length()) {
		variants.push_back(PrinsesVariant(pos, 0, newseq, thisVariant));
	}
}
#endif


/* Trims common sequence at beginning and end of a variant, adding it to the list
	prinsesFrag = assembled contig
	refseqFrag = corresponding refseq 
*/
static list<Variant>
findVariants(DnaSequence refseqFrag, DnaSequence prinsesFrag, char direction)
{
	list<Variant> variants;

	const GenomicPosition mfl = refseqFrag.length();
	const GenomicPosition pfl = prinsesFrag.length();
	GenomicPosition commonStart, commonEnd;		//number of bases in common

	if (direction == 'd') { // greedy commonStart
		for (commonStart = 0;
		commonStart < mfl and commonStart < pfl and refseqFrag[commonStart] == prinsesFrag[commonStart];
		++commonStart) ;
		for (commonEnd = 1;
		commonEnd <= mfl - commonStart and commonEnd <= pfl - commonStart and refseqFrag[mfl-commonEnd] == prinsesFrag[pfl-commonEnd];
		++commonEnd);
		--commonEnd;
	} else {		// greedy commonEnd
		for (commonEnd = 1;
		commonEnd <= mfl and commonEnd <= pfl and refseqFrag[mfl-commonEnd] == prinsesFrag[pfl-commonEnd];
		++commonEnd);
		--commonEnd;
		for (commonStart = 0;
		commonStart < mfl - commonEnd and commonStart < pfl - commonEnd and refseqFrag[commonStart] == prinsesFrag[commonStart];
		++commonStart);
	}

	if (commonStart) {
		refseqFrag = refseqFrag.substr(commonStart);
		prinsesFrag = prinsesFrag.substr(commonStart);
	}
	if (commonEnd) {
		refseqFrag = refseqFrag.substr(0, -(int32_t) commonEnd);
		prinsesFrag = prinsesFrag.substr(0, -(int32_t) commonEnd);
	}

	Variant::findVariants(refseqFrag, prinsesFrag, variants, commonStart);
	return variants;
}


class Job : public ThreadedJob
{
public:
	typedef enum { Queued, Running, Finished, Processed } Status;

	int _jobIndex;

	uint16_t _chromoIndex;
	Strand _strand;

	Prinses::AssemblyParameters _params;
	const Genome& _genome;
	const String& _outDirectory;
	Status _status;		// this is only changed by the scheduler (not mutexed)
	const char* const _refseq;
	const GenomicPosition _refseqLength;
	const SamPileup& _sam;
//	const uint8_t* const _alignmentReadDepth;

	// returned:
	DnaNamedSequence _prinsesSequence;
	//list<PrinsesVariant> _variants;
	list<Variant> _variants;
	Uint32_t _numFinishedContigs;
	Uint32_t _numFailedLowCover;
	Uint32_t _numFailedMaxLength;
	Uint32_t _numFailedRepeat;
	Uint32_t _numMatchReference;
	Uint32_t _numReferenceRepeats;
	Uint32_t _numWithVariants;

	Job(int jobIndex, ChromosomeIndex chromo, Strand strand,
		const SamPileup& sam,
//		const uint8_t* readDepth,
		const Prinses::AssemblyParameters& params, const String& outDirectory,
		const Genome& genome)
		: _jobIndex(jobIndex), _chromoIndex(chromo), _strand(strand),
		_params(params), _genome(genome),
		_outDirectory(outDirectory), _status(Queued), _refseq(genome.chromoSequence(chromo, strand)),
		_refseqLength(genome.chromoLength(chromo)), _sam(sam) // _alignmentReadDepth(readDepth)
		{}
	virtual void run();
	void processVariants(ostream& variantsFile);
	void removeDuplicateVariants();

private:
	bool uniqueSequence(const char* s, GenomicPosition len, bool term = false) const;
};


bool
Job::uniqueSequence(const char* s, GenomicPosition len, bool term) const
{
	if (_strand.isPositive()) {
		const uint64_t a = _params._genomeFragmentsPositive.findFragment(s, len).numberRemaining();
		if (a > 1) return false;
		return a + _params._genomeFragmentsNegative.findFragment(s, len).numberRemaining() == 1;
	}
	const uint64_t a = _params._genomeFragmentsNegative.findFragment(s, len).numberRemaining();
	if (a > 1) return false;
	return a + _params._genomeFragmentsPositive.findFragment(s, len).numberRemaining() == 1;
}


// list of variants must be sorted
void
Job::removeDuplicateVariants()
{
	if (_variants.size() <= 1) return;

	list<Variant>::iterator thisVariantP = _variants.begin();
	list<Variant>::iterator lastVariantP = thisVariantP;
	list<Variant>::iterator prevVariantP = _variants.end(); // previous non-equal variant

	thisVariantP++;
	while (thisVariantP != _variants.end()) {
//	for (++thisVariantP; thisVariantP != _variants.end();
//	prevVariantP = lastVariantP, lastVariantP = thisVariantP, thisVariantP++)

		if ((*thisVariantP)._direction == 'b') continue;

		if (lastVariantP->_direction == thisVariantP->_direction
		or lastVariantP->_direction == 'b'
		or lastVariantP->oldSeqLength() != (*thisVariantP).oldSeqLength()
		or lastVariantP->newSeqLength() != (*thisVariantP).newSeqLength()) {
			prevVariantP = lastVariantP;
			lastVariantP = thisVariantP;
			thisVariantP++;
			continue;
		}

		int64_t posDiff = (int64_t) (*thisVariantP).position() - (int64_t) lastVariantP->position();
		DnaSequence lastSeq = lastVariantP->newSeq();
		if (posDiff)
			lastSeq += _genome.sequence(lastVariantP->position() + lastVariantP->oldSeqLength(), posDiff);
		DnaSequence thisSeq;
		if (posDiff) thisSeq = _genome.sequence(lastVariantP->position(), posDiff);
		thisSeq += (*thisVariantP).newSeq();
		if (lastSeq != thisSeq) {
			prevVariantP = lastVariantP;
			lastVariantP = thisVariantP;
			thisVariantP++;
			continue;
		}

		// now we know the two variants are the same
		// let's see if they overlap with previous variant:
		if (prevVariantP != _variants.end() and prevVariantP->position() + prevVariantP->oldSeqLength() >= lastVariantP->position()) {
			// remove variant at "lastVariantP"
//			thisVariantP->_contigIds.splice(thisVariantP->_contigIds.end(), lastVariantP->_contigIds);
			thisVariantP->_direction = 'b';
			_variants.erase(lastVariantP);
			lastVariantP = thisVariantP;
			thisVariantP++;
		} else {
			// remove variant at "thisVariantP"
//			lastVariantP->_contigIds.splice(lastVariantP->_contigIds.begin(), thisVariantP->_contigIds);
			lastVariantP->_direction = 'b';
			_variants.erase(thisVariantP);
			thisVariantP = lastVariantP;
			thisVariantP++;
		}
	}
}


void
Job::run()
{
	const time_t time0 = time(NULL);

	const char direction = _strand.isPositive() ? 'd' : 'u';
//	const uint8_t *alignmentReadDepth = _alignmentReadDepth;
	const SamPileup& sam = _sam;
	Prinses::AssemblyParameters& params = _params;
	const char* const refseq = _refseq;
	const GenomicPosition refSeqLength = _refseqLength;
	const GenomicPosition refSeqEnd = refSeqLength - 1;
	int pctDone = 0;
	uint32_t contigNumber = 0;
	const String& seqName(_genome.chromoName(_chromoIndex));
	const GenomicPosition chromoStart = _genome.chromoStart(_chromoIndex);
	Locus loc;
	Vector<const char*>& terminators = params._terminators;

	ofstream logFile;
	if (verbose) {
		ios_base::openmode logFileMode = ios_base::out;
		String logFileName;
		if (maxThreads == 1) {
			stringstream ss;
			ss << _outDirectory << "/log.txt";
			logFileName = ss.str();
			logFileMode |= ios_base::app;
		} else {
			stringstream ss;
			ss << _outDirectory << "/" << seqName << "." << direction << ".log";
			logFileName = ss.str();
		}
		logFile.open(logFileName.c_str(), logFileMode);
		if (logFile.fail()) {
			cerr << programName << ": Can't open file " << logFileName << " for writing"
	#ifdef DEBUG
				<< " in " << __PRETTY_FUNCTION__ << " " << __FILE__ << " line " << __LINE__ 
	#endif
				<< endl;
			exit(1);
		}
		logFile << ">" << seqName << " direction " << direction << endl;
	}
//ostream& cdbg = logFile;

	{
		MutexLock ml(cerrMutex);
		cerr << endl << Util::getLocalTime(time0) << " starting " << seqName << " direction " << direction << endl;
	}

//#define DEPTH(pos) alignmentReadDepth[direction == 'd' ? (pos) : refSeqEnd - (pos)]
#define DEPTH(pos) sam.depthAtPosition((direction == 'd' ? (pos) : refSeqEnd - (pos)) + chromoStart)
	// first map the reads to the reference sequence:
//	float *coverage;
//	Prinses::mapReads(_refseq.c_str(), _refseq.length(), params._reads, genomeFragments, minMatchLength, coverage, &logFile);

	GenomicPosition refPos = primerLength;
	DnaNamedSequence prinsesSeq(seqName.c_str(), refseq, refPos);
	GenomicPosition lastTerPos = 0;

//	while (refPos + genomeFragmentLength <= refSeqLength) {
	while (refPos + primerLength <= refSeqLength) {
		if (showProgress and ((int) (refPos * 100 / refSeqLength)) > pctDone) {
			pctDone = (int) (refPos * 100 / refSeqLength);
			MutexLock ml(cerrMutex);
			cerr << ' ' << seqName << ':' << direction << ':' << pctDone << '%' << flush;
			if (pctDone % 10 == 0) cerr << endl;
		}

//		float rd = coverage[refPos];
//		if (//rd >= readDepthHigh or
//		 !( rd < readDepthLow
//		or rd + 15.0 < coverage[refPos - 5]
//		or rd + 20.0 < coverage[refPos - 10]
//		or rd < coverage[refPos - 5] * .85
//		or rd < coverage[refPos - 10] * .8)) {
//			prinsesSeq += refseq[refPos];
//			refPos++;
//			continue;
		float rd = DEPTH(refPos);
		if (rd >= readDepthHigh or
		 !( rd < readDepthLow
		or rd + 15.0 < DEPTH(refPos - 5)
		or rd + 20.0 < DEPTH(refPos - 10)
		or rd < DEPTH(refPos - 5) * .85
		or rd < DEPTH(refPos - 10) * .8)) {
			prinsesSeq += refseq[refPos];
			refPos++;
			continue;
		}

		params._primer = prinsesSeq.substr(-primerLength);
		if (!uniqueSequence(params._primer.c_str(), primerLength)) { // non-unique primer
			if (verbose) logFile << "non-unique primer at pos " << refPos + 1 << endl;
			prinsesSeq += refseq[refPos];
			refPos++;
			continue;
		}
		params._assemblyId = seqName;
		params._assemblyId += ':';
		params._assemblyId += direction;
		params._assemblyId += ':';
		params._assemblyId += refPos;
/* qqqq
		if (_nonUniqueShortSequences) {
			while (!frag->set(params._primer.c_str())
			or _nonUniqueShortSequences->contains(frag)) {
				if (verbose)
					logFile << "primer non-unique at " << (refPos+1) << endl;
				prinsesSeq += refseq[refPos];
				refPos++;
				if (refPos == refSeqLength) break;
				params._primer = prinsesSeq.substr(-primerLength);
			}
		}
*/

		// Now find suitable terminator:
		const GenomicPosition primerPos = refPos;		// the position JUST AFTER the primer
		if (verbose)
			logFile << "trying at " << primerPos+1 << " primer " << params._primer << " (depth " << rd << " vs " << DEPTH(refPos-5) << ", " << DEPTH(refPos-10) << ") " << flush;
		GenomicPosition terPos;
		float trd = 0;
		for (terPos = primerPos + 3; terPos < refSeqLength - terminatorLength and terPos < refPos + maxTerminatorDistance; ++terPos) {
			if (terPos == lastTerPos) continue;
//			trd = coverage[terPos];
//			if (trd < readDepthLow) continue;
//			if (trd >= coverage[terPos-5] + 15.0
//			or trd >= coverage[terPos-10] + 20.0
//			or trd >= coverage[terPos-5] * 1.15
//			or trd >= coverage[terPos-10] * 1.2) break;
			trd = DEPTH(terPos);
			if (trd < readDepthLow) continue;
			if (trd >= DEPTH(terPos-5) + 15.0
			or trd >= DEPTH(terPos-10) + 20.0
			or trd >= DEPTH(terPos-5) * 1.15
			or trd >= DEPTH(terPos-10) * 1.2) break;
		}
//		cdbg << "primerPos = " << primerPos << " termPos = " << terPos << " rd = " << rd << " trd = " << trd << endl;

		lastTerPos = terPos;

		// if terminator is at the end of the sequence:
		if (terPos >= refSeqLength - terminatorLength) {
			prinsesSeq += refseq[refPos];
			refPos++;
			if (verbose)
				logFile << " failed - can't find terminator " << endl;
			continue;
		}

		GenomicPosition allTerPos[numTerminators];
		GenomicPosition step = terminatorLength;
		terminators.set(NULL);
		uint16_t numTerm = 0;
		for (uint16_t i = 0; i < numTerminators; ++i, terPos += step) {
			if (terPos >= refSeqLength - terminatorLength or terPos >= refPos + maxTerminatorDistance)
				break;

			// make sure terminator is unique in genome:
			if (uniqueSequence(refseq + terPos, terminatorLength, true)) {
				allTerPos[i] = terPos;
				terminators[i] = refseq + terPos;
				numTerm++;
			} else {
				allTerPos[i] = 0;
			}
//			cdbg << "Term: refpos:" << refPos << " terpos:+" << (terPos - refPos) << endl;
			if (i == 0 or i == numTerminators - 1) {
			} else if (i == 1) {
#define TERM_BASE	4
				int n = numTerminators - 2;	// remaining num of terms
				int denom = 1, a = TERM_BASE;
				for (int j = 1; j < n; ++j, a *= TERM_BASE) denom += a;
				step = (refPos + maxTerminatorDistance - terPos) / denom;
			} else {
				step *= TERM_BASE;
			}
		}

		if (numTerm == 0) {
			if (verbose)
				logFile << "failed (no suitable terminator)" << endl;
			prinsesSeq += refseq[refPos];
			refPos++;
			continue;
		}

		params._maxAssemblyLength = allTerPos[0] - refPos + maxInsertionLength + primerLength;
		if (verbose)
			logFile << " for " << (allTerPos[0] - primerPos) << " (" << numTerm << " terminators)" << flush;
//			logFile << "trying at " << (direction == 'd' ? primerPos+1 : refSeqLength - primerPos) << " primer " << params._primer << " (depth " << rd << " vs " << coverage[refPos-5] << ", " << coverage[refPos-10] << ") for " << (allTerPos[0] - primerPos) << " ";

/*
			stringstream ss;
			ss << seqName << ":" << (direction == 'd' ? primerPos : refSeqEnd - primerPos);
			params._seqName = ss.str();
*/

		// qqq new addition, should be documented:
		if (minDepth == 0) {
			params._minDepth = (uint16_t) (rd / 4.0);
			if (params._minDepth < 1) params._minDepth = 1;
			params._maxDepth = maxDepth > params._minDepth ? maxDepth : params._minDepth;
		}

		list<Prinses::AssemblyResult> results;
		Prinses::assemble(params, results);
		contigNumber++;
		uint32_t versionNumber = 0;

		if (results.size() == 0) {
			if (verbose)
				logFile << " failed nr" << endl;
			prinsesSeq += refseq[refPos];
			refPos++;
			continue;
		}
//		const Prinses::AssemblyResult& tmp = *(results.begin());
//		cdbg << "assemble done: " << tmp._assemblyEndReason << ":" << tmp._contig.length() << endl;

		list<Prinses::AssemblyResult>::iterator resI;
		bool resultAppended = false;
		GenomicPosition contigLength = 0;

		// for each result (more than one results possible with branching)
		for (resI = results.begin(); resI != results.end(); ++resI) {
			Prinses::AssemblyResult& result = *resI;
			if (result._assemblyEndReason != Prinses::END_TERMINATOR_FOUND) {
				switch (result._assemblyEndReason) {
				case Prinses::END_LOW_COVER: _numFailedLowCover++; break;
				case Prinses::END_MAX_LENGTH: _numFailedMaxLength++; break;
				case Prinses::END_REPEATS: _numFailedRepeat++; break;
				case Prinses::END_GENOME_REPEAT: _numFailedRepeat++; break;
				case Prinses::END_TERMINATOR_FOUND: break; // gets rid of warning
				}
				continue;
			}
			_numFinishedContigs++;

			// "contig" does not contain the primer or the terminator:
			const DnaSequence& contig = result._contig;
			contigLength = contig.length();

			// termPos points to the beginning of the terminator
			const GenomicPosition termPos = allTerPos[result._terminatorIndex];
//			cdbg << "Term found at " << termPos << ": " << _refseq.substr(termPos, terminatorLength) << endl;

			const DnaSequence refFrag(refseq + primerPos, termPos - primerPos);

			// if we only have one result and it matches the reference, move on:
			if (refFrag == contig and results.size() == 1) {
				_numMatchReference++;
				if (verbose) logFile << " success " << contigLength << " - no change" << endl;
				if (!resultAppended) {
					prinsesSeq += refFrag;
					refPos += refFrag.length();
					resultAppended = true;
				}
				continue;
			}

/*
			// align reads and compare read depth before and after:
			GenomicPosition extraLeft;
			if (primerPos >= (GenomicPosition) primerLength + 10)
				extraLeft = 10;
			else
				extraLeft = primerPos - primerLength;

			GenomicPosition extraRight;
			if (termPos + terminatorLength + 10 < refSeqLength)
				extraRight = 10;
			else
				extraRight = refSeqLength - termPos - terminatorLength;

			DnaSequence contigPlus(refseq + primerPos - primerLength - extraLeft, primerLength + extraLeft);
			contigPlus += contig;
			contigPlus += _refseq.substr(termPos, primerLength + extraRight);
			float *coverageAfter;
			Prinses::mapReads(contigPlus.c_str(), contigPlus.length(), params._reads, genomeFragments, minMatchLength, coverageAfter, &logFile);
			GenomicPosition posBefore = primerPos - extraLeft, posAfter = primerLength;
			float covBefore = 0.0, covAfter = 0.0;
			for (GenomicPosition l = contigLength + extraLeft + extraRight; l; l--, posAfter++)
				covAfter += coverageAfter[posAfter];
			covAfter /= (contigLength + extraLeft + extraRight);
			for (GenomicPosition l = refFrag.length() + extraLeft + extraRight; l; l--, posBefore++)
				covBefore += coverage[posBefore];
			covBefore /= (refFrag.length() + extraLeft + extraRight);
			delete [] coverageAfter;

			if (covAfter < covBefore + 1.0) {
				if (verbose) logFile << " failed - coverage does not increase (" << covBefore << "->" << covAfter << ")" << endl;
				if (!resultAppended) {
					refPos = termPos;
					prinsesSeq += contig;
					resultAppended = true;
				}
				continue;
			}
			*/

			if (!allowRepeatsInReference and refFrag.length() >= (GenomicPosition) 2 * primerLength) {
				ShortSequenceSortedArray* nonUnique = ShortSequenceSortedArray::nonUniqueShortSequences(refFrag, primerLength);
				if (nonUnique and nonUnique->number() > 0) {
					delete nonUnique;
					_numReferenceRepeats++;
					if (verbose)
						logFile << " failed - reference sequence repeats" << endl;
					continue;
				}
				if (nonUnique) delete nonUnique;
			}

			if (verbose)
				logFile << " success " << contigLength << endl;
//				logFile << " success " << contigLength << " - coverage increases from " << covBefore << " to " << covAfter << endl;

			const list<Variant> simpleVariants = findVariants(refFrag, contig, direction);
			list<Variant>::const_iterator vi;
			MutexLock l(variantsRawFileMutex);
			if (simpleVariants.size()) _numWithVariants++;
			for (vi = simpleVariants.begin(); vi != simpleVariants.end(); ++vi) {
				GenomicPosition chromoPos;
				DnaSequence newSeq;
				if (direction == 'd') {
					chromoPos = vi->position() + primerPos;
					newSeq = vi->newSeq();
				} else {
					chromoPos = refSeqLength - vi->oldSeqLength() - vi->position() - primerPos;
					newSeq = vi->newSeq().reverseComplement();
				}
				const GenomicPosition pos = _genome.position(_chromoIndex, chromoPos);
				_variants.push_back(Variant(pos, vi->oldSeqLength(), newSeq, direction));
				if (variantsRawFile)
					_genome.output(*variantsRawFile, *(_variants.rbegin()));

/*				_variants.push_back(PrinsesVariant(pos, vi->oldSeqLength(), newSeq, direction,
					contigNumber, versionNumber, resI->_score / (contigLength || 1), primerPos, contigLength,
					(uint16_t) covBefore, (uint16_t) covAfter));
*/				if (verbose) _genome.output(logFile, *(_variants.rbegin()));
			}
			versionNumber++;
			if (!resultAppended) {
				refPos = termPos;
				prinsesSeq += contig;
				resultAppended = true;
			}
		}

		if (!resultAppended) {
			if (verbose) logFile << " failed " << contigLength << endl;
			prinsesSeq += refseq[refPos];
			refPos++;
		}
	}
	if (refPos < refSeqLength)
		prinsesSeq += String(refseq + refPos, refSeqLength - refPos);

	if (direction == 'd') {
		_prinsesSequence = prinsesSeq;
	} else {
		_prinsesSequence.set(prinsesSeq.reverseComplement());
		_prinsesSequence.setName(prinsesSeq.name());
	}

	MutexLock ml(cerrMutex);
	cerr << endl << seqName << ":" << direction << " done, " << _variants.size() << " unprocessed variants found in " << Util::hms(time(NULL)-time0) << endl;
	if (verbose)
		logFile << seqName << " direction " << direction << " " << _variants.size() << " unprocessed variants found" << endl << endl;

//	delete [] coverage;
}


// This takes a pair of Job containing the two directions of results outputs results
void
Job::processVariants(ostream& variantsFile)
{
	list<Variant>::const_iterator it;
	_variants.sort();

	removeDuplicateVariants();

	// output variants:
	const list <Variant>& variants(_variants);
	int numSNPs = 0;
	int numInsertions = 0;
	int numDeletions = 0;
	int numOther = 0;
	list<int> insLength, delLength;
	for (it = variants.begin(); it != variants.end(); ++it) {
		GenomicPosition nsl = (*it).newSeqLength();
		GenomicPosition osl = (*it).oldSeqLength();
		if (nsl == 1 and osl == 1) {
			numSNPs++;
		} else if (nsl == 0) {
			numDeletions++;
			delLength.push_back(osl);
		} else if (osl == 0) {
			numInsertions++;
			insLength.push_back(nsl);
		} else {
			numOther++;
		}
		_genome.output(variantsFile, *it);
	}
	cerr << "\rSequence " << _genome.chromoName(_chromoIndex)  << ": total " << variants.size() << " variants, of which "
	<<  numSNPs << " SNPs, " << numInsertions << " insertions, " << numDeletions << " deletions, "
	<< numOther << " other variants" << endl;
	if (insLength.size()) {
		insLength.sort();
		GenomicPosition i, size = insLength.size();
		list<int>::iterator it;
		cerr << "Insertion length median ";
		for (it = insLength.begin(), i=0; it != insLength.end(); ++it, ++i) {
			if (i == size/2)
				cerr << (*it) << ", max ";
			if (i == size-1)
				cerr << (*it) << endl;
		}
	}
	if (delLength.size()) {
		delLength.sort();
		GenomicPosition i, size = delLength.size();
		list<int>::iterator it;
		cerr << "Deletion length median ";
		for (it = delLength.begin(), i=0; it != delLength.end(); ++it, ++i) {
			if (i == size/2)
				cerr << (*it) << ", max ";
			if (i == size-1)
				cerr << (*it) << endl;
		}
	}
}


void
main(int argc, char* argv[])
{
	SimpleAllocator::initialise();
	Util::setThousandSeparator(cout);
	Util::setThousandSeparator(cerr);

	stringstream signature;
	signature << argv[0] << ' ' << argv[1];
	programName = signature.str();
	argc--; argv++;
	for (int i = 1; i < argc; ++i)
		signature << ' ' << argv[i];
#ifdef DEBUG
	signature << " debug";
#endif

	time_t time0 = time(NULL);
	cerr << Util::getLocalTime(time0) << " " << signature.str() << endl;
	signal(SIGTERM, signalHandler);
//	signal(SIGSEGV, sigsegHandler);

	// Get command line arguments:
	map<String, bool*> flags;
	map<String, int*> intOptions;
	map<String, String*> stringOptions;
	String viewFileName;
	flags["version"] = &printVersion;
	intOptions["minMatchLength"] = &minMatchLength;
	intOptions["minDepth"] = &minDepth;
	intOptions["maxDepth"] = &maxDepth;
	intOptions["readDepthLow"] = &readDepthLow;
	intOptions["readDepthHigh"] = &readDepthHigh;
	intOptions["rollingDepthLow"] = &rollingDepthLow;
	intOptions["maxInsertionLength"] = &maxInsertionLength;
	intOptions["minMappingQuality"] = &minMappingQuality;
	stringOptions["viewFile"] = &viewFileName;
	flags["verbose"] = &verbose;
	intOptions["maxTerminatorDistance"] = &maxTerminatorDistance;
	intOptions["maxThreads"] = &maxThreads;
	intOptions["terminatorLength"] = &terminatorLength;
	intOptions["primerLength"] = &primerLength;
	intOptions["numTerminators"] = &numTerminators;
//	intOptions["maxBranches"] = &maxBranches;
	flags["useScores"] = &useScores;
	flags["allowRepeatsInReference"] = &allowRepeatsInReference;
	flags["progress"] = &showProgress;
	vector<String> argV = Util::getCommandLineOptions(flags, intOptions, stringOptions, argc, argv, usage);
	if (printVersion) {
		cerr << "built on " << __DATE__ << " " << __TIME__ << endl;
		exit(0);
	}
	flags.erase("version");
	if (argV.size() != 4) usage();
	if (maxThreads <= 0) maxThreads = 1;
	if (maxThreads > 1 and viewFileName.length()) {
		cerr << programName << ": view file specified, must run single-threaded" << endl;
		maxThreads = 1;
	}
	if (minMappingQuality > 60) {
		cerr << programName << ": Warning: minMappingQuality too high, set to 60" << endl;
		minMappingQuality = 60;
	} else if (minMappingQuality < 0) {
		cerr << programName << ": Warning: minMappingQuality negative, set to 0" << endl;
		minMappingQuality = 0;
	}

	int arg = 0;

	// Command line arguments:
	String genomeFile = argV[arg++];
	String alignmentBamFile = argV[arg++];
	String readsFile = argV[arg++];
	String outDirectory = argV[arg++];

	// See if we can open all files first:
	ifstream readsFH(readsFile.c_str(), ios_base::in|ios_base::binary);
	if (readsFH.fail()) {
		cerr << programName << ": Can't open file " << readsFile 
#ifdef DEBUG
			<< " in " << __PRETTY_FUNCTION__ << " " << __FILE__ 
			<< " line " << __LINE__ 
#endif
			<< endl;
		exit(1);
	}

/*	ifstream maqViewFH(maqViewFile.c_str(), ios_base::in);
	if (maqViewFH.fail()) {
		cerr << programName << ": Can't open file " << maqViewFile
#ifdef DEBUG
			<< " in " << __PRETTY_FUNCTION__ << " " << __FILE__ 
			<< " line " << __LINE__ 
#endif
			<< endl;
		exit(1);
	}
*/
	Util::checkCreateDirectory(outDirectory.c_str());
	String variantsFileName = outDirectory + "/variants.txt";
	ofstream variantsFile(variantsFileName.c_str());
	if (variantsFile.fail()) {
		cerr << programName << ": Can't open file " << variantsFileName << " for writing"
#ifdef DEBUG
			<< " in " << __PRETTY_FUNCTION__ 
			<< " " << __FILE__ << " line " << __LINE__ 
#endif
			<< endl;
		exit(1);
	}

	cerr << Util::getLocalTime(time(NULL)) << " loading genome" << endl;
	Genome referenceGenome(genomeFile.c_str());
	SamPileup sam(alignmentBamFile.c_str(), referenceGenome, minDepth, minMappingQuality);

	if (viewFileName.length()) {
		viewFile = new ofstream(viewFileName.c_str());
		if (viewFile->fail()) {
			cerr << programName << ": Can't open file " << viewFileName 
#ifdef DEBUG
				<< " in " << __PRETTY_FUNCTION__ 
				<< " " << __FILE__ << " line " << __LINE__ 
#endif
				<< endl;
			exit(1);
		}
	}

	String variantsRawFileName = outDirectory + "/raw.variants.txt";
	variantsRawFile = new ofstream(variantsRawFileName.c_str());
	if (variantsRawFile->fail()) {
		cerr << programName << ": Can't open file " << variantsRawFileName 
#ifdef DEBUG
			<< " in " << __PRETTY_FUNCTION__ 
			<< " " << __FILE__ << " line " << __LINE__ 
#endif
			<< endl;
		exit(1);
	}

	String downstreamSequenceFileName = outDirectory + "/downstream.fa";
	ofstream downstreamSequenceFile(downstreamSequenceFileName.c_str());
	if (downstreamSequenceFile.fail()) {
		cerr << programName << ": Can't open file " << downstreamSequenceFileName << " for writing"
#ifdef DEBUG
			<< " in " << __PRETTY_FUNCTION__ 
			<< " " << __FILE__ << " line " << __LINE__ 
#endif
			<< endl;
		exit(1);
	}
	String upstreamSequenceFileName = outDirectory + "/upstream.fa";
	ofstream upstreamSequenceFile(upstreamSequenceFileName.c_str());
	if (upstreamSequenceFile.fail()) {
		cerr << programName << ": Can't open file " << upstreamSequenceFileName << " for writing"
#ifdef DEBUG
			<< " in " << __PRETTY_FUNCTION__ 
			<< " " << __FILE__ << " line " << __LINE__ 
#endif
			<< endl;
		exit(1);
	}

	String alignmentConsensusFileName = outDirectory + "/alignmentConsensus.fa";
	ofstream alignmentConsensusFile(alignmentConsensusFileName.c_str());
	if (alignmentConsensusFile.fail()) {
		cerr << programName << ": Can't open file " << alignmentConsensusFileName << " for writing"
#ifdef DEBUG
			<< " in " << __PRETTY_FUNCTION__ 
			<< " " << __FILE__ << " line " << __LINE__ 
#endif
			<< endl;
		exit(1);
	}

	if (maxThreads == 1) {
		String logFileName(outDirectory);
		logFileName += "/log.txt";
		remove(logFileName.c_str());
	}


	// Load maq profile:
//	map<String, const uint8_t*> alignmentReadDepth;
//	Vector<DnaNamedSequence*> alignmentConsensus;
//	uint16_t averageReadDepth = Prinses::loadMaqProfile(maqViewFH, alignmentConsensus, &alignmentReadDepth);
//	Genome genome(alignmentConsensus);
	const Genome& genome = sam.targetGenome();
	genome.output(String(outDirectory + "/alignmentConsensus.fa").c_str());
//	alignmentConsensus.freeAll();
//	alignmentConsensus.resize(0);

	double averageReadDepth = sam.averageReadDepth();
	if (!readDepthHigh) readDepthHigh = (int) (averageReadDepth + averageReadDepth/2.0);
//	maqViewFH.close();
	cerr << Util::getLocalTime(time(NULL)) << " loaded alignment profile" << endl;
/*
	bool fail = false;
	for (ChromosomeIndex i = 0; i < genome.numberChromosomes(); ++i) {
		map<String, const uint8_t*>::iterator it = alignmentReadDepth.find(genome.chromoName(i));
		if (it == alignmentReadDepth.end()) {
			cerr << programName << ": chromosome " << genome.chromoName(i) << " not found in alignment file " << maqViewFile << endl;
			fail = true;
		}
	}
	if (fail) exit(1);
*/

	// Load reads:
	ShortSequenceSortedArray* reads = ShortSequenceSortedArray::loadReads(readsFH);
	readsFH.close();
	if (!reads->hasScores()) useScores = false;
	if (reads->number() == 0 or reads->readLength() <= 16) {
		cerr << programName << ": No reads found or reads too short"
#ifdef DEBUG
			<< " in " << __PRETTY_FUNCTION__ 
			<< " " << __FILE__ << " line " << __LINE__ 
#endif
			<< endl;
		exit(1);
	}

	if (primerLength <= 0)
		primerLength = min(reads->readLength() - 1, 100);
	if (terminatorLength <= 0)
		terminatorLength = primerLength;

	const GenomeFragmentArray genomeFragmentsPositive(genome, Strand::positive, &cerr, terminatorLength);
	const GenomeFragmentArray genomeFragmentsNegative(genome, Strand::negative, &cerr, terminatorLength);


	cerr << Util::getLocalTime(time(NULL)) << " all files loaded" << endl;

	variantsFile << "# " << Util::getLocalTime(time0) << " '" << signature.str() << "'" << endl;
	variantsFile << "#" << setw(30) << "number of unique reads" << " = " << reads->number() << endl;
	variantsFile << "#" << setw(30) << "read length" << " = " << reads->readLength() << endl;
	variantsFile << "#" << setw(30) << "average read depth" << " = " << averageReadDepth << endl;
	{
		map<String, bool*>::iterator flagsIt;
		for (flagsIt = flags.begin(); flagsIt != flags.end(); ++flagsIt)
			variantsFile << "#" << setw(30) << flagsIt->first << " = " << (*(flagsIt->second) ? "true" : "false") << endl;
		map<String, int*>::iterator intOptionsIt;
		for (intOptionsIt = intOptions.begin(); intOptionsIt != intOptions.end(); ++intOptionsIt)
			variantsFile << "#" << setw(30) << intOptionsIt->first << " = " << *(intOptionsIt->second) << endl;
		map<String, String*>::iterator stringOptionsIt;
		for (stringOptionsIt = stringOptions.begin(); stringOptionsIt != stringOptions.end(); ++stringOptionsIt)
			variantsFile << "#" << setw(30) << stringOptionsIt->first << " = " << *(stringOptionsIt->second) << endl;
	}
//	variantsFile << "# SeqName\tPosition\tLength\tNew Sequence\tDirection.ContigNo.Version\tAverage Score" << endl;
	variantsFile << "# SeqName:direction\tPosition\tLength\tNew Sequence" << endl;
	Prinses::AssemblyParameters params(viewFile, verbose, useScores, maxBranches, *reads, 
		minDepth, maxDepth, minMatchLength,
		maxTerminatorDistance + maxInsertionLength + primerLength, numTerminators, terminatorLength,
		genomeFragmentsPositive, genomeFragmentsNegative);

	int numJobs = genome.numberChromosomes() * 2;
	Job* jobs[numJobs];

	if (maxThreads == 1) {
		for (ChromosomeIndex i = 0; i < genome.numberChromosomes(); ++i) {
//			const uint8_t* const rd = alignmentReadDepth[genome.chromoName(i)];
			jobs[2*i] = new Job(2*i, i, Strand::positive, sam, params, outDirectory, genome);
			jobs[2*i]->run();
			downstreamSequenceFile << jobs[2*i]->_prinsesSequence;

			jobs[2*i+1] = new Job(2*i+1, i, Strand::negative, sam, params, outDirectory, genome);
			jobs[2*i+1]->run();
			upstreamSequenceFile << jobs[2*i+1]->_prinsesSequence;

			jobs[2*i]->_variants.merge(jobs[2*i+1]->_variants);

			jobs[2*i]->processVariants(variantsFile);
		}
	} else {
		ThreadScheduler threadScheduler(maxThreads);
		for (ChromosomeIndex i = 0; i < genome.numberChromosomes(); ++i) {
//			const uint8_t* const rd = alignmentReadDepth[genome.chromoName(i)];
			threadScheduler.addJob(jobs[2*i] = new Job(2*i, i, Strand::positive, sam, params, outDirectory, genome));
			threadScheduler.addJob(jobs[2*i+1] = new Job(2*i+1, i, Strand::negative, sam, params, outDirectory, genome));
		}
		ThreadedJob* tjob;
		while ((tjob = threadScheduler.waitForNext()) != NULL) {
			Job& thisJob = * (Job*) tjob;
			thisJob._status = Job::Finished;

			// see if any pair of jobs is finished
			for (int j = 0; j < numJobs; j += 2) {
				if (jobs[j]->_status != Job::Finished or jobs[j+1]->_status != Job::Finished)
					continue;

				jobs[j]->_variants.merge(jobs[j+1]->_variants);

				jobs[j]->processVariants(variantsFile);

				jobs[j]->_status = Job::Processed;
				jobs[j+1]->_status = Job::Processed;
				break;
			}

			cerr << Util::getLocalTime(time(NULL)) << " job finished: " << genome.chromoName(thisJob._chromoIndex)
				 << ":" << (thisJob._strand.isPositive() ? 'd' : 'u') << endl;

			if (thisJob._strand.isPositive()) 
				downstreamSequenceFile << thisJob._prinsesSequence;
			else
				upstreamSequenceFile << thisJob._prinsesSequence;
		}

		for (int j = 0; j < numJobs; j += 2) {
			if (jobs[j]->_status != Job::Finished or jobs[j+1]->_status != Job::Finished)
				continue;

			jobs[j]->_variants.merge(jobs[j+1]->_variants);
			jobs[j]->processVariants(variantsFile);

			jobs[j]->_status = Job::Processed;
			jobs[j+1]->_status = Job::Processed;
		}
	}

	// job assembly stats:
	cerr << "Assembly stats (number of contigs):" << endl;
	for (ChromosomeIndex i = 0; i < genome.numberChromosomes(); ++i) {
		cerr << '\t' << genome.chromoName(i) << ":d\t" << genome.chromoName(i) << ":u";
	}
	cerr << "\ttotal" << endl;
	cerr << "finished";
	uint32_t total = 0;
	for (ChromosomeIndex i = 0; i < genome.numberChromosomes(); ++i) {
		cerr << '\t' << jobs[2*i]->_numFinishedContigs << '\t' << jobs[2*i+1]->_numFinishedContigs;
		total +=  jobs[2*i]->_numFinishedContigs + jobs[2*i+1]->_numFinishedContigs;
	}
	cerr << '\t' << total << endl;
	cerr << "match reference";
	total = 0;
	for (ChromosomeIndex i = 0; i < genome.numberChromosomes(); ++i) {
		cerr << '\t' << jobs[2*i]->_numMatchReference << '\t' << jobs[2*i+1]->_numMatchReference;
		total +=  jobs[2*i]->_numMatchReference + jobs[2*i+1]->_numMatchReference;
	}
	cerr << '\t' << total << endl;
	cerr << "reference repeats";
	total = 0;
	for (ChromosomeIndex i = 0; i < genome.numberChromosomes(); ++i) {
		cerr << '\t' << jobs[2*i]->_numReferenceRepeats << '\t' << jobs[2*i+1]->_numReferenceRepeats;
		total +=  jobs[2*i]->_numReferenceRepeats + jobs[2*i+1]->_numReferenceRepeats;
	}
	cerr << '\t' << total << endl;
	cerr << "with variants";
	total = 0;
	for (ChromosomeIndex i = 0; i < genome.numberChromosomes(); ++i) {
		cerr << '\t' << jobs[2*i]->_numWithVariants << '\t' << jobs[2*i+1]->_numWithVariants;
		total +=  jobs[2*i]->_numWithVariants + jobs[2*i+1]->_numWithVariants;
	}
	cerr << '\t' << total << endl;

	cerr << "failed (low cover)";
	total = 0;
	for (ChromosomeIndex i = 0; i < genome.numberChromosomes(); ++i) {
		cerr << '\t' << jobs[2*i]->_numFailedLowCover << '\t' << jobs[2*i+1]->_numFailedLowCover;
		total +=  jobs[2*i]->_numFailedLowCover + jobs[2*i+1]->_numFailedLowCover;
	}
	cerr << '\t' << total << endl;
	cerr << "failed (max length)";
	total = 0;
	for (ChromosomeIndex i = 0; i < genome.numberChromosomes(); ++i) {
		cerr << '\t' << jobs[2*i]->_numFailedMaxLength << '\t' << jobs[2*i+1]->_numFailedMaxLength;
		total +=  jobs[2*i]->_numFailedMaxLength + jobs[2*i+1]->_numFailedMaxLength;
	}
	cerr << '\t' << total << endl;
	cerr << "failed (repeat)";
	total = 0;
	for (ChromosomeIndex i = 0; i < genome.numberChromosomes(); ++i) {
		cerr << '\t' << jobs[2*i]->_numFailedRepeat << '\t' << jobs[2*i+1]->_numFailedRepeat;
		total +=  jobs[2*i]->_numFailedRepeat + jobs[2*i+1]->_numFailedRepeat;
	}
	cerr << '\t' << total << endl;

	// wind down:
	if (viewFile) delete viewFile;
	if (variantsRawFile) delete variantsRawFile;

	delete reads;
	for (int i = 0; i < numJobs; ++i) delete jobs[i];

	SimpleAllocator::freeAll();
	time_t time1 = time(NULL);
	cerr << endl << Util::getLocalTime(time1) << " " << programName << " done in " << Util::hms(time1-time0) << endl;
}

}; // namespare PrinsesG
